#ifndef GALIAS_H
#define GALIAS_H
#include "kwdset.h"


// aliases for element types
enum gtypel {noel=0, isolinear1d=1, isoquadratic1d=2, trianglelinear=3, trianglequadratic=4, isolinear2d=5, isoquadratic2d=6, 
             tetrahedronlinear=7, tetrahedronquadratic=8, pyramidelinear=9, pyramidequadratic=10, vedgelinear=11,
             vedgequadratic=12, isolinear3d=13, isoquadratic3d=14, all_element_types=15};
const enumstr gtypelstr[] = {{"noel",0}, {"isolinear1d", 1}, {"isoquadratic1d", 2}, {"trianglelinear", 3}, {"trianglequadratic", 4}, 
                             {"isolinear2d", 5}, {"isoquadratic2d", 6}, {"tetrahedronlinear", 7}, {"tetrahedronquadratic", 8}, 
                             {"pyramidelinear", 9}, {"pyramidequadratic", 10}, {"vedgelinear", 11}, {"vedgequadratic", 12},
                             {"isolinear3d", 13}, {"isoquadratic3d", 14}};
const kwdset gtypel_kwdset(sizeof(gtypelstr)/sizeof(*gtypelstr), gtypelstr);



// aliases for entity types in T3d
enum entityp {evertex=1, ecurve=2, esurface=3, eregion=4, epatch=5, eshell=6};
const enumstr entitypstr[] = {{"evertex", 1}, {"ecurve", 2}, {"esurface", 3}, {"eregion", 4}, {"epatch", 5}, {"eshell", 6}};
const kwdset entityp_kwdset(sizeof(entitypstr)/sizeof(*entitypstr), entitypstr);



// aliases for general enities
enum gentity {gvertex=1, gcurve=2, gsurface=3, gregion=4};



//  aliases for mesh description
enum meshdescription {all_nodes=1,bound_nodes=2,neg_bound_nodes=3,glob_glued=4,metis=11,metis_elem=12};
const enumstr meshdescriptionstr[] = {{"all_nodes",1}, {"bound_nodes",2}, {"neg_bound_nodes",3}, {"glob_glued",4},
				      {"metis",11}, {"metis_elem",12}};
const kwdset meshdescription_kwdset(sizeof(meshdescriptionstr)/sizeof(*meshdescriptionstr), meshdescriptionstr);



// aliases for matrix storage in memory
enum storagetype {without_matrix=0,
                  dense_matrix=1,skyline_matrix=2,double_skyline=3,
                  compressed_rows=10,symm_comp_rows=11,element_matrices=40,
                  lapack_stor=100,petsc=120,spdirect_stor_scr=140,spdirect_stor_cr=141};
const enumstr storagetypestr[] = {{"without_matrix",0}, {"dense_matrix",1}, {"skyline_matrix", 2},
                                  {"double_skyline",3}, {"compressed_rows",10},{"symm_comp_rows", 11},
                                  {"element_matrices", 40}, {"lapack_stor", 100}, {"petsc", 120}, 
                                  {"spdirect_stor_scr",140}, {"spdirect_stor_cr", 141}};
const kwdset storagetype_kwdset(sizeof(storagetypestr)/sizeof(*storagetypestr), storagetypestr);



// aliases for solvers of linear equation system
enum linsolvertype {no_solver=0,gauss_elim=1,ldl=2,lu=3,ll=4,ill=5,conden=11,cg=20,bicg=30,
                    lapack_sol=100,pardisolver=120,spdirldl=140,spdirlu=141,spdirll=142,
                    sschur=151,sfeti=161,sfetidef=171, saddle_point=181};
const enumstr linsolvertypestr[] = {{"gauss_elim",1}, {"ldl",2}, {"lu",3}, {"ll",4}, {"ill",5}, {"conden",11},
                                    {"cg",20}, {"bicg",30}, {"lapack_sol",100}, {"pardisolver",120}, {"spdirldl",140},
                                    {"spdirlu",141}, {"spdirll",142},
				    {"sschur",151}, {"sfeti",161}, {"sfetidef",171},{"saddle_point",181}};
const kwdset linsolvertype_kwdset(sizeof(linsolvertypestr)/sizeof(*linsolvertypestr), linsolvertypestr);



//  aliases for type of FETI implementation
enum fetiimplem {no_impl=0,boolean_matrices=1,nonredundant=2,redundant=3};
const enumstr fetiimplemstr[] = {{"no_impl",0}, {"boolean_matrices",1}, {"nonredundant",2}, {"redundant",3}};
const kwdset fetiimplem_kwdset(sizeof(fetiimplemstr)/sizeof(*fetiimplemstr), fetiimplemstr);



//  aliases for type of solver of coarse problem
enum redsystsolver {master_sol=1,paral_sol=2};
const enumstr redsystsolverstr[] = {{"master_sol",1}, {"paral_sol",2}};
const kwdset redsystsolver_kwdset(sizeof(redsystsolverstr)/sizeof(*redsystsolverstr), redsystsolverstr);



// aliases for preconditioners for conjugate gradients
enum precondtype {noprecond=0,diagprec=1,ssorprec=5,incomdec=10,sparseindec=20,boss=101,feti_lumped=151,feti_dirichlet=152};
const enumstr precondtypestr[] = {{"noprecond",0}, {"diagprec",1}, {"ssorprec",5}, {"incomdec",10}, {"sparseindec",20}, {"boss",101},{"feti_lumped",151},{"feti_dirichlet",152}};
const kwdset precondtype_kwdset(sizeof(precondtypestr)/sizeof(*precondtypestr), precondtypestr);



// aliases for node renumbering
enum noderenumb {no_renumbering=0,cuthill_mckee=1,rev_cuthill_mckee=2,sloan=3};
const enumstr noderenumbstr[] = {{"no_renumbering",0}, {"cuthill_mckee",1}, {"rev_cuthill_mckee",2}, {"sloan",3}};
const kwdset noderenumb_kwdset(sizeof(noderenumbstr)/sizeof(*noderenumbstr), noderenumbstr);



// aliases for eigenvalues solver
enum eigensolver {no_eigsolver=0,inv_iteration=1,subspace_it_jacobi=4,subspace_it_gsortho=5};
const enumstr eigensolverstr[] = {{"no_eigsolver",0}, {"inv_iteration",1}, {"subspace_it_jacobi",4}, {"subspace_it_gsortho",5}};
const kwdset eigensolver_kwdset(sizeof(eigensolverstr)/sizeof(*eigensolverstr), eigensolverstr);



// aliases for different types of general function definition
enum generalfunct {stat=0,pars=1,tab=2,pars_set=3,multpv=10,itab=20,gfunc_set=30};
const enumstr generalfunctstr[] = {{"stat",0}, {"pars",1}, {"tab",2}, {"pars_set",3}, {"multpv", 10}, {"itab",20}, {"gfunc_set",30}};
const kwdset generalfunct_kwdset(sizeof(generalfunctstr)/sizeof(*generalfunctstr), generalfunctstr);



// aliases for different types of selection
enum seltype {sel_no=0, sel_all=1, sel_range=2, sel_list=3, sel_period=4, 
              sel_realrange=5, sel_reallist=6, sel_mtx=7, sel_range_mtx=8, sel_range_vec=9,
              sel_realperiod=10, sel_impvalues=11};
const enumstr seltypestr[] = {{"sel_no",0}, {"sel_all",1}, {"sel_range",2}, {"sel_list",3}, 
                              {"sel_period",4}, {"sel_realrange",5}, {"sel_reallist",6}, 
			      {"sel_mtx",7}, {"sel_rangem",8}, {"sel_rangev",9}, 
                              {"sel_realperiod",10}, {"sel_impvalues",11}};
const kwdset seltype_kwdset(sizeof(seltypestr)/sizeof(*seltypestr), seltypestr);



// aliases for point specification
enum nodip {no_point=0, atnode=1, atip=2, atxyz=3};
const enumstr nodipstr[] = {{"no_point",0}, {"atnode",1}, {"atip",2}, {"atxyz",3}};
const kwdset nodip_kwdset(sizeof(nodipstr)/sizeof(*nodipstr), nodipstr);



// aliases for simple flag
enum flagsw {off=0, on=1};
const enumstr flagswstr[] = {{"off",0}, {"on",1}};
const kwdset flagsw_kwdset(sizeof(flagswstr)/sizeof(*flagswstr), flagswstr);



// aliases for simple answer type
enum answertype {no=0, yes=1};
const enumstr answertypestr[] = {{"no",0}, {"yes",1}};
const kwdset answertype_kwdset(sizeof(answertypestr)/sizeof(*answertypestr), answertypestr);



// aliases for header/footer flag
enum hflagsw {hf_off=0, header=1, footer=2};
const enumstr hflagswstr[] = {{"hf_off",0}, {"header",1}, {"footer",2}};
const kwdset hflagsw_kwdset(sizeof(hflagswstr)/sizeof(*hflagswstr), hflagswstr);

// aliases for general types of finite elements
enum gelemtype {noelem=0,
		linbar=1, quadbar=2,
                lintriag=21,quadtriag=22,linquad=25,quadquad=26,
                lintetra=41,quadtetra=42,linhexa=45,quadhexa=46,gen2d=50};
const enumstr gelemtypestr[] = {{"noelem",0},
				{"linbar",1},{"quadbar",2},
				{"lintriag",21},{"quadtriag",22},{"linquad",25},{"quadquad",26},
				{"lintetra",41},{"quadtetra",42},{"linhexa",45},{"quadhexa",46}, 
                                {"gen2d",50}};
const kwdset gelemtype_kwdset(sizeof(gelemtypestr)/sizeof(*gelemtypestr),gelemtypestr);



//  aliases for mesh format
enum meshform {sifel=0, t3d=1};
const enumstr meshformstr[] = {{"sifel",0}, {"t3d",1}};
const kwdset meshform_kwdset(sizeof(meshformstr)/sizeof(*meshformstr), meshformstr);



//  aliases for type of time controller
enum timecontrtype {fixed=0, adaptive=1, adaptivemin=2, adaptivemax=3};
const enumstr timecontrtypestr[] = {{"fixed",0}, {"adaptive",1}, {"adaptivemin",2}, {"adaptivemax",3}};
const kwdset timecontrtype_kwdset(sizeof(timecontrtypestr)/sizeof(*timecontrtypestr), timecontrtypestr);


// aliases for harddisk backup type
enum hdbackuptype {nohdb=0, hdbs_single=1, hdbs_multiple=2, hdbr_single=3, hdbr_multiple=4, hdbrs_single=5, 
                   hdbrs_multiple=6, hdbr_nonloc=7, hdbs_nonloc=8, hdbrs_nonloc=9};
const enumstr hdbackuptypestr[] = {{"nohdb",0}, {"hdbs_single",1}, {"hdbs_multiple",2}, {"hdbr_single",3}, {"hdbr_multiple",4}, 
                                   {"hdbrs_single",5}, {"hdbrs_multiple",6}, {"hdbr_nonloc",7}, {"hdbs_nonloc",8}, 
                                   {"hdbrs_nonloc",9}};
const kwdset hdbackuptype_kwdset(sizeof(hdbackuptypestr)/sizeof(*hdbackuptypestr), hdbackuptypestr);


// aliases for harddisk backup file format
enum hdbackupfmttype {text=1, binary=2};
const enumstr hdbackupfmttypestr[] = {{"text",1}, {"binary",2}};
const kwdset hdbackupfmttype_kwdset(sizeof(hdbackupfmttypestr)/sizeof(*hdbackupfmttypestr), hdbackupfmttypestr);


// aliases for keyword processing switch
enum pkwd_sw {nokwd=0, kwd_pd=1, kwd_outd=2, kwd_pd_outd=3, kwd_full=4};


// aliases for edge types used in material interfaces, jumps, hemivariational inequalities, etc.
enum edgetype {noedge=0,jumps=1, mater=2};
const enumstr edgetypestr[] = {{"noedge",0}, {"jumps",1}, {"mater",2}};
const kwdset edgetype_kwdset(sizeof(edgetypestr)/sizeof(*edgetypestr), edgetypestr);

#endif


 
